/*
 * Colors.h - Basic color functions
 *
 * Concept and Implementation by: Craig A. Lindley
 * Version: 1.1
 * Last Update: 03/15/2015
 */

#ifndef COLORS_H
#define COLORS_H

#define NUM_OF_COLOR_VALUES 256
#define MIN_COLOR_VALUE       0
#define MAX_COLOR_VALUE     255

// HSV to RGB color conversion
// Input arguments
// hue in degrees (0 - 360.0)
// saturation (0.0 - 1.0)
// value (0.0 - 1.0)
// Output arguments
// red, green blue (0.0 - 1.0)
void hsvToRGB(float hue, float saturation, float value, float *red, float *green, float *blue) {

  int i;
  float f, p, q, t;

  if (saturation == 0) {
    // achromatic (grey)
    *red = *green = *blue = value;
    return;
  }
  hue /= 60;                  // sector 0 to 5
  i = floor(hue);
  f = hue - i;                // factorial part of h
  p = value * (1 - saturation);
  q = value * (1 - saturation * f);
  t = value * (1 - saturation * (1 - f));
  switch (i) {
    case 0:
      *red = value;
      *green = t;
      *blue = p;
      break;
    case 1:
      *red = q;
      *green = value;
      *blue = p;
      break;
    case 2:
      *red = p;
      *green = value;
      *blue = t;
      break;
    case 3:
      *red = p;
      *green = q;
      *blue = value;
      break;
    case 4:
      *red = t;
      *green = p;
      *blue = value;
      break;
    default:
      *red = value;
      *green = p;
      *blue = q;
      break;
  }
}


// Create a HSV color and return it to an RGB structure
void createColor(float hue, float saturation, float value, CRGB *color) {
  float r, g, b;

  hsvToRGB(hue, saturation, value, &r, &g, &b);
  (*color).red   = (uint8_t) (r * MAX_COLOR_VALUE);
  (*color).green = (uint8_t) (g * MAX_COLOR_VALUE);
  (*color).blue =  (uint8_t) (b * MAX_COLOR_VALUE);
}

// Create an HSV color
void createHSVColor(int divisions, int index, float saturation, float value, CRGB *color) {

  float hueAngle = (360.0 * index) / divisions;
  createColor(hueAngle, saturation, value, color);
}

// Create a fully saturated HSV color
void createHSVColor(int divisions, int index, CRGB *color) {
  index %= divisions;
  createHSVColor(divisions, index, 1.0, 1.0, color);
}

/*******************************************************************/
/***             Palette definitions and Functions               ***/
/*******************************************************************/

#define PALETTE_SIZE      256
#define NUM_OF_PALETTES     7
#define GRAYSCALE_PALETTE   0
#define SPECTRUM_PALETTE    1
#define SIN1_PALETTE        2
#define SIN2_PALETTE        3
#define SIN3_PALETTE        4
#define SIN4_PALETTE        5
#define RANDOM_PALETTE      6

// Create a palette structure for holding color information
CRGB palette[PALETTE_SIZE];

// Generate a palette based upon parameter
void generatePalette(int paletteNumber) {

  // Create some factors for randomizing the generated palettes
  // This helps keep the display colors interesting
  float f1 = random16(16, 128);
  float f2 = random16(16, 128);
  float f3 = random16(16, 128);

  switch (paletteNumber) {
    case GRAYSCALE_PALETTE:
      {
        // Grayscale palette
        // Sometimes the light colors at low index; other times at high index
        boolean direction = (random16(2) == 1);
        if (direction) {
          for (int i = 0; i < PALETTE_SIZE; i++) {
            palette[i].red = palette[i].green = palette[i].blue = i;
          }
        }
        else  {
          for (int i = 0; i < PALETTE_SIZE; i++) {
            palette[i].red = palette[i].green = palette[i].blue = 255 - i;
          }
        }
      }
      break;

    case SPECTRUM_PALETTE:
      {
        // Full spectrum palette at full saturation and value
        for (int i = 0; i < PALETTE_SIZE; i++) {
          createHSVColor(PALETTE_SIZE, i, &palette[i]);
        }
      }
      break;

    case SIN1_PALETTE:
      {
        // Use sin function to generate palette
        for (int i = 0; i < PALETTE_SIZE; i++) {
          palette[i].red   = 128.0 + 128 * sin(PI * i / f1);
          palette[i].green = 128.0 + 128 * sin(PI * i / f2);
          palette[i].blue  = 128.0 + 128 * sin(PI * i / f3);
        }
      }
      break;

    case SIN2_PALETTE:
      {
        // Use sin function to generate palette - no blue
        for (int i = 0; i < 256; i++) {
          palette[i].red   = 128.0 + 128 * sin(PI * i / f1);
          palette[i].green = 128.0 + 128 * sin(PI * i / f2);
          palette[i].blue  = 0;
        }
      }
      break;

    case SIN3_PALETTE:
      {
        // Use sin function to generate palette - no green
        for (int i = 0; i < 256; i++) {
          palette[i].red   = 128.0 + 128 * sin(PI * i / f1);
          palette[i].green = 0;
          palette[i].blue  = 128.0 + 128 * sin(PI * i / f2);
        }
      }
      break;

    case SIN4_PALETTE:
      {
        // Use sin function to generate palette - no red
        for (int i = 0; i < PALETTE_SIZE; i++) {
          palette[i].red   = 0;
          palette[i].green = 128.0 + 128 * sin(PI * i / f1);
          palette[i].blue  = 128.0 + 128 * sin(PI * i / f2);
        }
      }
      break;

    case RANDOM_PALETTE:
      {
        // Choose random color components
        for (int i = 0; i < PALETTE_SIZE; i++) {
          palette[i].red   = random16(NUM_OF_COLOR_VALUES);
          palette[i].green = random16(NUM_OF_COLOR_VALUES);
          palette[i].blue  = random16(NUM_OF_COLOR_VALUES);
        }
      }
      break;
  }
}

#endif

