/*
 * JuliaSet.h - Julia Set Display Pattern
 *
 * Concept and Implementation by: Craig A. Lindley
 * Version: 1.1
 * Last Update: 03/15/2015
 */

#ifndef JULIASET_H
#define JULIASET_H

#include "Necklace.h"

#define MANDELBROT_STEPS 256

class JuliaSet :
  public PatternBase {

  public:
    void setup() {
      terminated = false;
    }

    void drawPattern() {
      if (terminated) {
        return;
      }

      lcd.clearScreen();

      // Generate a complex constant as seed for Julia set
      float cReal = generateRandomFloat();
      float cImag = generateRandomFloat();

      // Make sure this complex constant diverges at (0, 0)
      // This prevents some not so interesting Julia sets from being rendered
      while (pointInMandelbrotSet(0, 0, cReal, cImag) == MANDELBROT_STEPS) {
        cReal = generateRandomFloat();
        cImag = generateRandomFloat();
      }
      // Draw the hopefully interesting Julia set
      drawJuliaSet(random16(NUM_OF_PALETTES), cReal, cImag);

      delay(1000);
    }

  private:
    // Private member functions
    void shufflePalette();
    int pointInMandelbrotSet(float z0Real, float z0Imag, float cReal, float cImag);
    float generateRandomFloat();
    void drawJuliaSet(int paletteNumber, float cReal, float cImag);
};

void JuliaSet::shufflePalette() {

  CRGB temp;

  for (int i = 0; i < PALETTE_SIZE; i++) {
    int r = random16(i, PALETTE_SIZE);
    temp = palette[r];
    palette[i] = palette[r];
    palette[r] = temp;
  }
}

int JuliaSet::pointInMandelbrotSet(float z0Real, float z0Imag, float cReal, float cImag) {

  float x = z0Real;
  float y = z0Imag;
  float nx, ny;
  for (int i = 0; i < MANDELBROT_STEPS; i++) {

    // Calculate the real part of the sequence
    nx = (x * x) - (y * y) + cReal;
    // Calculate the imaginary part of the sequence
    ny = 2 * x * y + cImag;
    // Check magnitude at each step
    // We check if it's greater than 2
    // which means the point diverges towards infinity
    if (((nx * nx) + (ny * ny)) > 4) {
      return i;
    }
    x = nx;
    y = ny;
  }
  // Point in mandelbrot set
  return MANDELBROT_STEPS;
}

// Generate a random signed 10 digit floating point value
float JuliaSet::generateRandomFloat() {

  char str[12];

  memset(str, 0, sizeof(str));
  int index = 0;

  boolean isPositive = (random16(2) == 1);
  if (isPositive) {
    str[index] = ' ';
  }
  else  {
    str[index] = '-';
  }
  index++;
  str[index++] = '.';

  for (int i = 0; i < 9; i++) {
    int offset = random16(10);
    str[index++] = '0' + offset;
  }
  return atof(str);
}

// Draw julia set
void JuliaSet::drawJuliaSet(int paletteNumber, float cReal, float cImag) {

  // Generate specified palette
  generatePalette(paletteNumber);

  shufflePalette();

  float xOfs = DISPLAY_WIDTH  / 2.0;
  float yOfs = DISPLAY_HEIGHT / 2.0;

  for (int x = 0; x < DISPLAY_WIDTH; x++) {
    float newReal = (((float) x - xOfs) / xOfs);

    for (int y = 0; y < DISPLAY_HEIGHT; y++) {
      float newImag = (((float) y - yOfs) / yOfs);

      int iterations = pointInMandelbrotSet(newReal, newImag, cReal, cImag);
      if (iterations != MANDELBROT_STEPS) {

        // Retrieve and convert color from current palette
        uint16_t color = lcd.Color565(palette[iterations].red, palette[iterations].green, palette[iterations].blue);
        lcd.drawPixel(x, y, color);
      }
    }
  }
}

#endif

