/*
 * Mandala.h - Mandala display patterns
 *
 * Implementation based on code written by: Stefan Petrick, 2014
 *
 * Adaptation and Implementation by: Craig A. Lindley
 * Version: 1.1
 * Last Update: 03/15/2015
 */

#ifndef MANDALA_H
#define MANDALA_H

#include "Necklace.h"

// Noise parameters
const int NOISE_WIDTH        = 64;
const int NOISE_WIDTH_MINUS  = NOISE_WIDTH - 1;
const int NOISE_HEIGHT       = 64;
const int NOISE_HEIGHT_MINUS = NOISE_HEIGHT - 1;
const int NOISE_CENTER_X     = NOISE_WIDTH  / 2;
const int NOISE_CENTER_Y     = NOISE_HEIGHT / 2;

// Display offsets for each quad
const int QUAD_1_X_OFFSET = 0;
const int QUAD_1_Y_OFFSET = 0;
const int QUAD_2_X_OFFSET = NOISE_WIDTH;
const int QUAD_2_Y_OFFSET = 0;
const int QUAD_3_X_OFFSET = NOISE_WIDTH;
const int QUAD_3_Y_OFFSET = NOISE_HEIGHT;
const int QUAD_4_X_OFFSET = 0;
const int QUAD_4_Y_OFFSET = NOISE_HEIGHT;

class Mandala :
  public PatternBase {

  public:

    void setup() {

      x = y = z = 0;

      // Allocate storage for noise
      noise = (uint8_t *) calloc(NOISE_WIDTH * NOISE_HEIGHT, 1);
      if (! noise) {
        Serial.printf("Could not allocate noise array\n");
        while (true) {
        }
      }

      // Allocate storage for screen
      screen = (CRGB *) calloc(NOISE_WIDTH * NOISE_HEIGHT, sizeof(CRGB));
      if (! screen) {
        Serial.printf("Could not allocate screen array\n");
        while (true) {
        }
      }
      terminated = false;
    }

    void drawPattern() {

      if (terminated) {
        return;
      }

      EVERY_N_SECONDS(6) {

        createRandomPalette();

        dx = random16(4000) - 2000;
        dy = random16(4000) - 2000;
        dz = random16(4000) - 2000;
        scale_x = random16(5000) + 1000;
        scale_y = random16(5000) + 1000;
      }

      x += dx;
      y += dy;
      z += dz;

      // Fill array with noise
      fillNoise();

      // Render the noise into the screen buffer
      renderNoise();

      // Do the caleidoscope effect
      caleidoscope();

      delay(80);
    }

    void tearDown() {

      // Free allocated memory
      free(screen);
      free(noise);

      terminated = true;
    }


  private:
    // Private variables
    uint32_t x, y, z, scale_x, scale_y;
    int16_t dx, dy, dz;

    // Pointer to the array of noise values
    uint8_t *noise;

    // Pointer to array of screen values
    CRGB *screen;

    // Current color palette
    CRGBPalette16 palette;

    // Private member functions
    void setNoise(int x, int y, uint8_t noiseValue);
    uint8_t getNoise(int x, int y);
    void setScreen(int x, int y, CRGB color);
    CRGB getScreen(int x, int y);
    void createRandomPalette();
    void fillNoise();
    void renderNoise();
    void caleidoscope();
};

#endif

/****************************************************************/
/*                         Member Functions                     */
/****************************************************************/

void Mandala::setNoise(int x, int y, uint8_t noiseValue) {

  int index = (y * NOISE_WIDTH) + x;
  noise[index] = noiseValue;
}

uint8_t Mandala::getNoise(int x, int y) {

  int index = (y * NOISE_WIDTH) + x;
  return noise[index];
}

void Mandala::setScreen(int x, int y, CRGB color) {

  int index = (y * NOISE_WIDTH) + x;
  screen[index] = color;
}

CRGB Mandala::getScreen(int x, int y) {

  int index = (y * NOISE_WIDTH) + x;
  return screen[index];
}

// Create a random palette
void Mandala::createRandomPalette() {

  palette = CRGBPalette16(
              CHSV( random8(), 255, 0 ),
              CHSV( random8(), 255, 0),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 255),

              CHSV( random8(), 255, 0 ),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 0),

              CHSV( random8(), 255, 0 ),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 0),

              CHSV( random8(), 255, 255 ),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 0),
              CHSV( random8(), 255, 0));
}

// Fill the x/y array with 16-bit noise values
void Mandala::fillNoise() {

  int noisesmoothing = 128;

  for (register int i = 0; i < NOISE_WIDTH; i++) {
    uint32_t ioffset = scale_x * (i - NOISE_CENTER_X);
    for (register int j = 0; j < NOISE_HEIGHT; j++) {
      uint32_t joffset = scale_y * (j - NOISE_CENTER_Y);
      byte data = inoise16(x + ioffset, y + joffset, z) >> 8;
      setNoise(i, j, scale8(getNoise(i, j), noisesmoothing) + scale8(data, 256 - noisesmoothing));
    }
  }
}

// Render the noise into the screen buffer
void Mandala::renderNoise() {

  for (register int y = 0; y < NOISE_HEIGHT; y++) {
    for (register int x = 0; x < NOISE_WIDTH; x++) {
      // Assign a color depending on the actual palette
      setScreen(x, y, ColorFromPalette(palette, getNoise(x, y), 255));
    }
  }
}

// The caleidoscope function works directly within the display buffer.
// Rotates the first quadrant 3 times onto display with each rotation being 90 degrees
void Mandala::caleidoscope() {

  CRGB colorXY;
  CRGB colorYX;

  for (register int y = 0; y < NOISE_HEIGHT; y++) {
    for (register int x = 0; x < NOISE_WIDTH; x++) {

      // Read colors from screen buffer
      colorXY = getScreen(x, y);
      colorYX = getScreen(y, x);

      // Write colors into quad1
      lcd.drawPixel(QUAD_1_X_OFFSET + x, y, lcd.Color565(colorXY.red, colorXY.green, colorXY.blue));

      // Write colors into quad2
      lcd.drawPixel(QUAD_2_X_OFFSET + NOISE_WIDTH_MINUS - x, y, lcd.Color565(colorYX.red, colorYX.green, colorYX.blue));

      // Write colors into quad3
      lcd.drawPixel(QUAD_3_X_OFFSET + NOISE_WIDTH_MINUS - x, QUAD_3_Y_OFFSET + NOISE_HEIGHT_MINUS - y, lcd.Color565(colorXY.red, colorXY.green, colorXY.blue));

      // Write colors into quad4
      lcd.drawPixel(QUAD_4_X_OFFSET + x, QUAD_4_Y_OFFSET + NOISE_HEIGHT_MINUS - y, lcd.Color565(colorYX.red, colorYX.green, colorYX.blue));
    }
  }
}





