/*
 * Contemplator Necklace
 *
 * This program will run in a necklace I'm planning on making
 * It runs various display patterns repeatedly.
 * A pattern will be displayed for a programmable duration and then
 * the next pattern will be selected and run.
 *
 * This code requires the TFT_ILI9163C driver and FastLED libraries
 *
 * Concept and Implementation: Craig A. Lindley
 * Version: 1.1
 * Last Update: 03/15/2015
 */

#include <Adafruit_GFX.h>
#include <TFT_ILI9163C.h>
#include <SPI.h>

#include "FastLED.h"
#include "Colors.h"
#include "PatternBase.h"
#include "Cube.h"
#include "JuliaSet.h"
#include "Mandala.h"
#include "Noise.h"
#include "Peace.h"
#include "Plasma.h"
#include "Watercolor.h"
#include "Waves.h"
#include "Web.h"

#define PATTERN_DISPLAY_TIME_SECONDS 60
#define PATTERN_DISPLAY_TIME_MS      (PATTERN_DISPLAY_TIME_SECONDS * 1000)

// Define Teensy 3.1 to LCD hardware connections
#define CS_PIN       10
#define DC_PIN        9
#define RESET_PIN    14
#define BACKLIGHT_PIN 3

// Instantiate the display pattern objects
Cube cube;
Mandala mandala;
Noise noise;
Peace peace;
Plasma plasma;
Watercolor waterColor;
Waves waves;
Web web;
JuliaSet julia;

uint8_t currentPatternIndex;

// Array of all patterns
PatternBase *patterns[] = {
  /* &julia, */ &cube, &noise, &web, &waves, &mandala, &peace, &waterColor, &plasma
};

#define NUMBER_OF_PATTERNS (sizeof(patterns) / sizeof(PatternBase *))

// Pointer to selected display pattern
PatternBase *ptrPattern;

unsigned long patternChangeTime;

// Declare instance of display driver
TFT_ILI9163C lcd = TFT_ILI9163C(CS_PIN, DC_PIN, RESET_PIN);

/************************************************************************/
/*                             Program Setup                            */
/************************************************************************/

void setup() {

  lcd.begin();
  lcd.clearScreen();

  // Turn on backlight at full brightness
  pinMode(BACKLIGHT_PIN, OUTPUT);
  digitalWrite(BACKLIGHT_PIN, HIGH);

  // Initialize random number generator
  randomSeed(analogRead(A9));

  // Add entropy
  random16_add_entropy(random());

  // Setup pattern selection index
  currentPatternIndex = random16(NUMBER_OF_PATTERNS);

  // Time in the future when pattern should change
  patternChangeTime = 0;

  // Initialize pointer to NULL
  ptrPattern = NULL;
}

/************************************************************************/
/*                              Main Program                            */
/************************************************************************/

void loop() {

  // Add entropy each loop
  random16_add_entropy(random());

  // Is the pattern display time up ?
  if (millis() > patternChangeTime) {
    // Display time is up

    // Tear down display pattern if necessary
    if (ptrPattern != NULL) {
      ptrPattern->tearDown();
    }
    // Pick the next display pattern
    ptrPattern = patterns[currentPatternIndex++];

    // Clear the screen in preparation
    lcd.clearScreen();

    // Setup next display pattern
    ptrPattern->setup();

    // Do some house keeping
    if (currentPatternIndex >= NUMBER_OF_PATTERNS) {
      currentPatternIndex = 0;
    }
    // Calculate next pattern display time
    patternChangeTime = millis() + PATTERN_DISPLAY_TIME_MS;
  }
  else    {
    // Display time not yet up so render a frame of the pattern
    ptrPattern->drawPattern();
  }
}









