/*
 * Noise.h - Noise display patterns
 *
 * Implementation based on code written by: Jason Coon and Mark Kriegsman
 *
 * Adaptation and Implementation by: Craig A. Lindley
 * Version: 1.1
 * Last Update: 03/15/2015
 */

#ifndef NOISE_H
#define NOISE_H

#include "Necklace.h"

class Noise :
  public PatternBase {

  public:

    void setup() {

      // Create a random palette for display
      createRandomPalette();

      // Initialize our coordinates to some random values
      x = random16();
      y = random16();
      z = random16();

      speed = random16(3, 25);
      scale = random16(4, 13);

      terminated = false;
    }

    void drawPattern() {
      if (terminated) {
        return;
      }
      EVERY_N_SECONDS(8) {
        createRandomPalette();
      }
      drawNoise(speed, scale);
    }

    void tearDown() {
      terminated = true;
    }

  private:
    // Private data
    uint16_t x;
    uint16_t y;
    uint16_t z;

    // Current color palette
    CRGBPalette16 palette;

    uint16_t speed;
    uint16_t scale;

    // Private member functions
    void createRandomPalette();
    void setNoise(int x, int y, uint8_t noiseValue);
    uint8_t getNoise(int x, int y);
    void drawNoise(uint16_t speed, uint16_t scale);
};

// Create a random palette
void Noise::createRandomPalette() {

  palette = CRGBPalette16(
              CHSV( random8(), 255, 0 ),
              CHSV( random8(), 255, 0),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 255),

              CHSV( random8(), 255, 0 ),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 0),

              CHSV( random8(), 255, 0 ),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 0),

              CHSV( random8(), 255, 255 ),
              CHSV( random8(), 255, 255),
              CHSV( random8(), 255, 0),
              CHSV( random8(), 255, 0));
}

// speed - how fast time moves forward
// scale - how far apart the pixels in our noise matrix are
void Noise::drawNoise(uint16_t speed, uint16_t scale) {

  // We're using the x/y dimensions to map to the x/y pixels on the matrix.  We'll
  // use the z-axis for "time".  speed determines how fast time moves forward.  Try
  // 1 for a very slow moving effect, or 60 for something that ends up looking like
  // water.
  // uint16_t speed = 1; // almost looks like a painting, moves very slowly
  // uint16_t speed = 4; // nice and slow plasma
  // uint16_t speed = 20; // 20 is a nice starting speed, mixes well with a scale of 100
  // uint16_t speed = 33;
  // uint16_t speed = 100; // wicked fast!

  // Scale determines how far apart the pixels in our noise matrix are.  Try
  // changing these values around to see how it affects the motion of the display.  The
  // higher the value of scale, the more "zoomed out" the noise iwll be.  A value
  // of 1 will be so zoomed in, you'll mostly see solid colors.

  // uint16_t scale = 1; // mostly just solid colors
  // uint16_t scale = 10; // nice and big plasma
  // uint16_t scale = 4011; // very zoomed out and shimmery
  // uint16_t scale = 10; // 311

  // Fill the x/y array of 8-bit noise values using the inoise8 function.
  for (int i = 0; i < DISPLAY_WIDTH; i++) {
    int ioffset = scale * i;
    for (int j = 0; j < DISPLAY_HEIGHT; j++) {
      int joffset = scale * j;
      CRGB color = ColorFromPalette(palette, inoise8(x + ioffset, y + joffset, z), 255);
      lcd.drawPixel(i, j, lcd.Color565(color.red, color.green, color.blue));
    }
  }
  x += speed;
  y += speed;
  z += speed;
}


#endif












