/*
 * Watercolor.h - Watercolor like display patterns
 *
 * Portions of this code are adapted from Aurora: Copyright (c) 2014 Jason Coon
 * Portions of this code are adapted from "Rainbow Smoke" by Jozsef Fejes
 *
 * Adaptation and Implementation by: Craig A. Lindley
 * Version: 1.1
 * Last Update: 03/15/2015
 */

#ifndef WATERCOLOR_H
#define WATERCOLOR_H

#define SURFACE_WIDTH  64
#define SURFACE_HEIGHT 64

#define NUMCOLORS      16
#define COLOR_COUNT  4096

struct Point {
  int x, y;
};

class Watercolor :
  public PatternBase {

  public:

    void setup() {
      currentColorIndex = 0;
      terminated = false;

      // Allocate dynamic memory for the various arrays
      paletteArray     = (CRGB *)    calloc(COLOR_COUNT, sizeof(CRGB));
      surfaceArray     = (CRGB *)    calloc(SURFACE_WIDTH * SURFACE_HEIGHT, sizeof(CRGB));
      hasColorArray    = (boolean *) calloc(SURFACE_WIDTH * SURFACE_HEIGHT, sizeof(boolean));
      isAvailableArray = (boolean *) calloc(SURFACE_WIDTH * SURFACE_HEIGHT, sizeof(boolean));
      if ((paletteArray == 0) || (surfaceArray == 0) || (hasColorArray == 0) || (isAvailableArray == 0)) {
        Serial.printf("Could not allocate array memory\n");
        while (true) {
        }
      }
    }

    void drawPattern();

    void tearDown() {

      // Free dynamic memory
      free(paletteArray);
      free(surfaceArray);
      free(hasColorArray);
      free(isAvailableArray);

      terminated = true;
    }

  private:


    // Private data
    CRGB *paletteArray;
    CRGB *surfaceArray;
    boolean *hasColorArray;
    boolean *isAvailableArray;

    int currentColorIndex;
    int algorithm;

    // Private member functions
    CRGB createCRGBFromHSV(uint8_t hue, uint8_t sat, uint8_t val);
    void createPaletteRGB();
    void createPaletteGBR();
    void createPaletteBRG();
    void createPaletteHSV();
    void shuffleColors();
    void createPalette();
    int colorDifference(CRGB c1, CRGB c2);
    void markAvailableNeighbors(Point point);
    Point getAvailablePointWithClosestNeighborColor(CRGB color);
    Point getAvailablePointWithClosestAverageNeighborColor(CRGB color);
    Point getAvailablePoint(int algorithm, CRGB color);
};

// Create CRGB color from uint8_t hsv values
CRGB Watercolor::createCRGBFromHSV(uint8_t hue, uint8_t sat, uint8_t val) {
  float redF, greenF, blueF;
  CRGB color;

  // Convert byte values to float in range 0.0 .. 1.0
  float hueF = ((float) hue) / 255.0;
  float satF = ((float) sat) / 255.0;
  float valF = ((float) val) / 255.0;

  hsvToRGB(hueF, satF, valF, &redF, &greenF, &blueF);

  color.red   = redF   * 255;
  color.green = greenF * 255;
  color.blue  = blueF  * 255;

  return color;
}

// Create an RGB palette
void Watercolor::createPaletteRGB() {
  CRGB color;
  int i = 0;

  for (int b = 0; b < NUMCOLORS; b++) {
    for (int g = 0; g < NUMCOLORS; g++) {
      for (int r = 0; r < NUMCOLORS; r++) {
        color.red =   r * 255 / (NUMCOLORS - 1);
        color.green = g * 255 / (NUMCOLORS - 1);
        color.blue =  b * 255 / (NUMCOLORS - 1);
        paletteArray[i] = color;

        i++;
        if (i == COLOR_COUNT)
          return;
      }
    }
  }
}

// Create an GBR palette
void Watercolor::createPaletteGBR() {
  CRGB color;
  int i = 0;

  for (int r = 0; r < NUMCOLORS; r++) {
    for (int b = 0; b < NUMCOLORS; b++) {
      for (int g = 0; g < NUMCOLORS; g++) {
        color.red =   r * 255 / (NUMCOLORS - 1);
        color.green = g * 255 / (NUMCOLORS - 1);
        color.blue =  b * 255 / (NUMCOLORS - 1);
        paletteArray[i] = color;

        i++;
        if (i == COLOR_COUNT)
          return;
      }
    }
  }
}

void Watercolor::createPaletteBRG() {
  CRGB color;
  int i = 0;

  for (int r = 0; r < NUMCOLORS; r++) {
    for (int g = 0; g < NUMCOLORS; g++) {
      for (int b = 0; b < NUMCOLORS; b++) {
        color.red =   r * 255 / (NUMCOLORS - 1);
        color.green = g * 255 / (NUMCOLORS - 1);
        color.blue =  b * 255 / (NUMCOLORS - 1);
        paletteArray[i] = color;

        i++;
        if (i == COLOR_COUNT)
          return;
      }
    }
  }
}

void Watercolor::createPaletteHSV() {
  int i = 0;

  uint8_t startHue = random16(0, 256);

  for (uint8_t h = startHue; i < 1024; h += 8) {
    for (uint16_t s = 0; s < 256; s += 16) {
      if (i < COLOR_COUNT) {
        paletteArray[i] = createCRGBFromHSV(h, s, 255);
      }
      i++;
    }

    for (uint16_t v = 256; v > 0; v -= 16) {
      if (i < COLOR_COUNT) {
        paletteArray[i] = createCRGBFromHSV(h, 255, v);
      }
      i++;
    }
  }
}

// Shuffle the colors in the current palette
void Watercolor::shuffleColors() {
  for (int a = 0; a < COLOR_COUNT; a++) {
    int r = random16(a, COLOR_COUNT);
    CRGB temp = paletteArray[a];
    paletteArray[a] = paletteArray[r];
    paletteArray[r] = temp;
  }
}

// Create a randomly selected palette
void Watercolor::createPalette() {

  switch (random16(4)) {
    case 0:
      createPaletteRGB();
      shuffleColors();
      break;
    case 1:
      createPaletteGBR();
      shuffleColors();
      break;
    case 2:
      createPaletteBRG();
      shuffleColors();
      break;
    case 3:
      createPaletteHSV();
      break;
  }
}

// Calc difference in colors
int Watercolor::colorDifference(CRGB c1, CRGB c2) {

  int rDiff = c1.red   - c2.red;
  int gDiff = c1.green - c2.green;
  int bDiff = c1.blue  - c2.blue;

  return (rDiff * rDiff) + (gDiff * gDiff) + (bDiff * bDiff);
}

void Watercolor::markAvailableNeighbors(Point point) {

  for (int dy = -1; dy <= 1; dy++) {
    int ny = point.y + dy;

    if (ny == -1 || ny == SURFACE_HEIGHT)
      continue;

    for (int dx = -1; dx <= 1; dx++) {
      if (dx == 0 && dy == 0)
        continue;

      int nx = point.x + dx;

      if (nx == -1 || nx == SURFACE_WIDTH)
        continue;

      if (! hasColorArray[ny * SURFACE_WIDTH + nx]) {
        isAvailableArray[ny * SURFACE_WIDTH + nx] = true;
      }
    }
  }
}

Point Watercolor::getAvailablePointWithClosestNeighborColor(CRGB color) {
  Point best;

  // Find the pixel with the smallest difference between the current color and all of it's neighbors' colors
  int smallestDifference = 999999;
  for (int y = 0; y < SURFACE_HEIGHT; y++) {
    for (int x = 0; x < SURFACE_WIDTH; x++) {
      // Skip any that arent' available
      if (! isAvailableArray[y * SURFACE_WIDTH + x])
        continue;

      // Loop through its neighbors
      int smallestDifferenceAmongNeighbors = 999999;
      for (int dy = -1; dy <= 1; dy++) {
        if (y + dy == -1 || y + dy == SURFACE_HEIGHT)
          continue;

        for (int dx = -1; dx <= 1; dx++) {
          if (x == 0 && y == 0)
            continue;

          if (x + dx == -1 || x + dx == SURFACE_WIDTH)
            continue;

          int nx = x + dx;
          int ny = y + dy;

          // Skip any neighbors that don't already have a color
          if (! hasColorArray[ny * SURFACE_WIDTH + nx])
            continue;

          // Get surface value
          CRGB neighborColor = surfaceArray[ny * SURFACE_WIDTH + nx];

          int difference = colorDifference(neighborColor, color);
          if (difference < smallestDifferenceAmongNeighbors || (difference == smallestDifferenceAmongNeighbors && random16(2) == 1)) {
            smallestDifferenceAmongNeighbors = difference;
          }
        }
      }

      if (smallestDifferenceAmongNeighbors < smallestDifference || (smallestDifferenceAmongNeighbors == smallestDifference && random16(2) == 1)) {
        smallestDifference = smallestDifferenceAmongNeighbors;
        best.x = x;
        best.y = y;
      }
    }
  }
  return best;
}

Point Watercolor::getAvailablePointWithClosestAverageNeighborColor(CRGB color) {
  Point best;

  int smallestAverageDifference = 999999;

  for (int y = 0; y < SURFACE_HEIGHT; y++) {
    for (int x = 0; x < SURFACE_WIDTH; x++) {
      // Skip any that aren't available
      if (! isAvailableArray[y * SURFACE_WIDTH + x])
        continue;

      int neighborCount = 0;
      int neighborColorDifferenceTotal = 0;

      // Loop through its neighbors
      for (int dy = -1; dy <= 1; dy++) {
        if (y + dy == -1 || y + dy == SURFACE_HEIGHT)
          continue;

        for (int dx = -1; dx <= 1; dx++) {
          if (x + dx == -1 || x + dx == SURFACE_WIDTH)
            continue;

          int nx = x + dx;
          int ny = y + dy;

          // skip any neighbors that don't already have a color
          if (! hasColorArray[ny * SURFACE_WIDTH + nx])
            continue;

          neighborCount++;

          CRGB neighborColor = surfaceArray[ny * SURFACE_WIDTH + nx];

          int difference = colorDifference(neighborColor, color);
          neighborColorDifferenceTotal += difference;
        }
      }

      int averageDifferenceAmongNeighbors = neighborColorDifferenceTotal / neighborCount;

      if (averageDifferenceAmongNeighbors < smallestAverageDifference || (averageDifferenceAmongNeighbors == smallestAverageDifference && random16(2) == 1)) {
        smallestAverageDifference = averageDifferenceAmongNeighbors;
        best.x = x;
        best.y = y;
      }
    }
  }
  return best;
}

Point Watercolor::getAvailablePoint(int algorithm, CRGB color) {
  if (algorithm == 0) {
    return getAvailablePointWithClosestNeighborColor(color);
  }  else  {
    return getAvailablePointWithClosestAverageNeighborColor(color);
  }
}

void Watercolor::drawPattern() {

  Point point;

  do {
    if (terminated) {
      return;
    }

    if (currentColorIndex == 0) {

      // Clear the surface
      memset(surfaceArray, 0, SURFACE_WIDTH * SURFACE_HEIGHT * sizeof(CRGB));

      lcd.clearScreen();
      createPalette();
      algorithm = random16(2);

      // clear all flags
      for (int y = 0; y < SURFACE_HEIGHT; y++) {
        for (int x = 0; x < SURFACE_WIDTH; x++) {
          hasColorArray[y * SURFACE_WIDTH + x] = false;
          isAvailableArray[y * SURFACE_WIDTH + x] = false;
        }
      }
    }
    // Get color from the palette
    CRGB color = paletteArray[currentColorIndex];

    if (currentColorIndex == 0) {
      // Pick a random starting point
      point.x = random16(SURFACE_WIDTH);
      point.y = random16(SURFACE_HEIGHT);
    }
    else    {
      point = getAvailablePoint(algorithm, color);
    }

    isAvailableArray[point.y * SURFACE_WIDTH + point.x] = false;
    hasColorArray[point.y * SURFACE_WIDTH + point.x] = true;

    // Store color onto surface
    surfaceArray[point.y * SURFACE_WIDTH + point.x] = color;

    uint16_t pixelColor = lcd.Color565(color.red, color.green, color.blue);

    // Expand pixel into 2x2 block to cover whole display
    int x = 2 * point.x;
    int y = 2 * point.y;

    lcd.drawPixel(x    , y,     pixelColor);
    lcd.drawPixel(x + 1, y,     pixelColor);
    lcd.drawPixel(x    , y + 1, pixelColor);
    lcd.drawPixel(x + 1, y + 1, pixelColor);

    markAvailableNeighbors(point);

    currentColorIndex++;
  }
  while (currentColorIndex < COLOR_COUNT);

  currentColorIndex = 0;

  delay(1500);
}

#endif




